/*
    Copyright (C) 2011 jeanfi@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "list.h"
#include "lp_ws.h"
#include "ppastats.h"

static struct package_stats *get_package_stats(struct ppa_stats *stats,
					       const char *name)

{
	struct package_stats *p, **p_cur;

	p_cur = stats->packages;
	while (p_cur && *p_cur) {
		struct package_stats *p = *p_cur;

		if (!strcmp(p->name, name))
			return p;

		p_cur++;
	}

	p = malloc(sizeof(struct package_stats));
	p->name = strdup(name);
	p->versions = NULL;
	p->download_count = 0;
	p->daily_download_totals = NULL;

	stats->packages = (struct package_stats **)list_add
		((void **)stats->packages, p);

	return p;
}

static struct version_stats *get_version_stats(struct package_stats *package,
					       const char *version)
{
	struct version_stats *v, **cur;

	cur = package->versions;
	while (cur && *cur) {
		struct version_stats *v = *cur;

		if (!strcmp(v->version, version))
			return v;

		cur++;
	}

	v = malloc(sizeof(struct version_stats));
	v->version = strdup(version);
	v->distros = NULL;
	v->download_count = 0;
	v->daily_download_totals = NULL;

	package->versions
		= (struct version_stats **)list_add((void **)package->versions,
						    v);

	return v;
}

static struct distro_stats *get_distro_stats(struct version_stats *version,
					     const char *name)
{
	struct distro_stats **cur = version->distros;
	struct distro_stats *d;

	while (cur && *cur) {
		d = *cur;

		if (!strcmp(d->name, name))
			return d;

		cur++;
	}

	d = malloc(sizeof(struct distro_stats));
	d->name = strdup(name);
	d->archs = NULL;
	d->download_count = 0;
	d->daily_download_totals = NULL;

	version->distros
		= (struct distro_stats **)list_add((void **)version->distros,
						   d);

	return d;
}

static struct arch_stats *get_arch_stats(struct distro_stats *distro,
					 const char *name)
{
	struct arch_stats **cur = distro->archs;
	struct arch_stats *a;

	while (cur && *cur) {
		a = *cur;

		if (!strcmp(a->name, name))
			return a;

		cur++;
	}

	a = malloc(sizeof(struct arch_stats));
	a->name = strdup(name);
	a->daily_download_totals = NULL;
	a->download_count = 0;

	distro->archs
		= (struct arch_stats **)list_add((void **)distro->archs,
						 a);

	return a;
}


static struct daily_download_total **add_total
(struct daily_download_total **totals, struct daily_download_total *total)
{
	struct daily_download_total **cur;
	struct daily_download_total *item;

	if (totals) {
		cur = totals;
		while (*cur) {
			item = *cur;

			if (item->date.tm_year == total->date.tm_year &&
			    item->date.tm_mon == total->date.tm_mon &&
			    item->date.tm_mday == total->date.tm_mday) {
				item->count += total->count;
				return totals;
			}

			cur++;
		}
	}

	item = malloc(sizeof(struct daily_download_total));
	memcpy(item, total, sizeof(struct daily_download_total));

	return (struct daily_download_total **)
		list_add((void **)totals, (void *)item);
}

struct daily_download_total **add_totals
(struct daily_download_total **total1, struct daily_download_total **total2)
{
	struct daily_download_total **cur;
	struct daily_download_total **result;

	result = total1;
	cur = total2;
	while (*cur) {
		result = add_total(result, *cur);

		cur++;
	}

	return result;
}

struct ppa_stats *
create_ppa_stats(const char *owner,
		 const char *ppa,
		 const char *package_status)
{
	struct ppa_stats *ppastats;
	struct binary_package_publishing_history **history;
	struct binary_package_publishing_history **h_cur;
	struct binary_package_publishing_history *h;
	char *ppa_url, *package_name, *package_version;
	struct package_stats *package;
	struct version_stats *version;
	const struct distro_series *distro_series;
	const struct distro_arch_series *arch_series;
	struct distro_stats *distro;
	struct arch_stats *arch;
	int count;
	struct daily_download_total **totals;

	ppa_url = get_archive_url(owner, ppa);
	history = get_binary_package_publishing_history_list(ppa_url,
							     package_status);
	free(ppa_url);

	if (!history) {
		fprintf(stderr, "Failed to retrieve PPA information\n");
		exit(EXIT_FAILURE);
	}

	ppastats = malloc(sizeof(struct ppa_stats));
	ppastats->name = strdup(ppa);
	ppastats->owner = strdup(owner);
	ppastats->packages = NULL;
	ppastats->daily_download_totals = NULL;
	ppastats->download_count = 0;

	h_cur = history;
	while (*h_cur) {
		h = *h_cur;
		package_name = h->binary_package_name;
		package_version = h->binary_package_version;
		arch_series
			= get_distro_arch_series(h->distro_arch_series_link);
		distro_series
			= get_distro_series(arch_series->distroseries_link);

		count = get_download_count(h->self_link);

		package = get_package_stats(ppastats, package_name);
		package->download_count += count;

		version = get_version_stats(package, package_version);
		version->download_count += count;

		distro = get_distro_stats(version, distro_series->name);
		distro->download_count += count;

		arch = get_arch_stats(distro, arch_series->architecture_tag);
		arch->download_count += count;

		ppastats->download_count += count;

		totals = get_daily_download_totals(h->self_link);

		ppastats->daily_download_totals
			= add_totals(ppastats->daily_download_totals,
				     totals);

		package->daily_download_totals
			= add_totals(package->daily_download_totals,
				     totals);

		daily_download_total_list_free(totals);

		h_cur++;
	}

	binary_package_publishing_history_list_free(history);

	return ppastats;
}

static void arch_stats_free(struct arch_stats *arch)
{
	free(arch->name);
	free(arch);
}

static void distro_stats_free(struct distro_stats *distro)
{
	struct arch_stats **archs;

	archs = distro->archs;
	if (archs) {
		while (*archs) {
			arch_stats_free(*archs);
			archs++;
		}
		free(distro->archs);
	}

	free(distro->name);
	free(distro);
}

static void version_stats_free(struct version_stats *version)
{
	struct distro_stats **distros;

	distros = version->distros;
	if (distros) {
		while (*distros) {
			distro_stats_free(*distros);
			distros++;
		}
		free(version->distros);
	}

	free(version->version);
	free(version);
}

static void package_stats_free(struct package_stats *package)
{
	struct version_stats **versions;

	versions = package->versions;
	if (versions) {
		while (*versions) {
			version_stats_free(*versions);
			versions++;
		}
		free(package->versions);
	}

	free(package->name);
	free(package);
}

void ppa_stats_free(struct ppa_stats *ppastats)
{
	struct package_stats **packages;

	packages = ppastats->packages;
	if (packages) {
		while (*packages) {
			package_stats_free(*packages);
			packages++;
		}
		free(ppastats->packages);
	}

	free(ppastats->owner);
	free(ppastats->name);

	daily_download_total_list_free(ppastats->daily_download_totals);

	free(ppastats);
}
